@@ -23,7 +23,7 @@
 /*
  * TODO: - locking
  *       - cancellation
- *       - get rid of g_main_loop (bug 555436#c32)
+ *       - error handling
  */
 
 #include <config.h>
@@ -100,8 +100,7 @@ static gboolean avahi_initialized = FALSE;
 static void free_global_avahi_client (void);
 static AvahiClient *get_global_avahi_client (GError **error);
 
-static gboolean ensure_avahi_resolver (GVfsDnsSdResolver  *resolver,
-                                       GError            **error);
+static void ensure_avahi_resolver (GVfsDnsSdResolver  *resolver);
 
 static void service_resolver_cb (AvahiServiceResolver   *resolver,
                                  AvahiIfIndex            interface,
@@ -139,7 +138,7 @@ remove_client_from_resolver (GVfsDnsSdResolver *resolver)
 static void
 add_client_to_resolver (GVfsDnsSdResolver *resolver)
 {
-  ensure_avahi_resolver (resolver, NULL);
+  ensure_avahi_resolver (resolver);
 }
 
 /* Callback for state changes on the Client */
@@ -217,23 +216,13 @@ get_global_avahi_client (GError **error)
   return global_client;
 }
 
-
 static gboolean
-ensure_avahi_resolver (GVfsDnsSdResolver  *resolver,
-                       GError            **error)
+start_avahi_resolver (gpointer user_data)
 {
+  GVfsDnsSdResolver *resolver = G_VFS_DNS_SD_RESOLVER (user_data);
   AvahiClient *avahi_client;
-  gboolean ret;
 
-  ret = FALSE;
-
-  if (resolver->avahi_resolver != NULL)
-    {
-      ret = TRUE;
-      goto out;
-    }
-
-  avahi_client = get_global_avahi_client (error);
+  avahi_client = get_global_avahi_client (NULL);
   if (avahi_client == NULL)
     goto out;
 
@@ -247,20 +236,19 @@ ensure_avahi_resolver (GVfsDnsSdResolver  *resolver,
                                                          0, /* AvahiLookupFlags */
                                                          service_resolver_cb,
                                                          resolver);
-  if (resolver->avahi_resolver == NULL)
-    {
-      g_set_error (error,
-                   G_IO_ERROR,
-                   G_IO_ERROR_FAILED,
-                   _("Error creating Avahi resolver: %s"),
-                   avahi_strerror (avahi_client_errno (avahi_client)));
-      goto out;
-    }
-
-  ret = TRUE;
 
 out:
-  return ret;
+  g_object_unref (resolver);
+  return FALSE;
+}
+
+static void
+ensure_avahi_resolver (GVfsDnsSdResolver  *resolver)
+{
+  if (resolver->avahi_resolver != NULL)
+    return;
+
+  g_idle_add (start_avahi_resolver, g_object_ref (resolver));
 }
 
 static void
@@ -469,7 +457,7 @@ g_vfs_dns_sd_resolver_constructed (GObject *object)
                                                          resolver->domain);
 
   /* start resolving immediately */
-  ensure_avahi_resolver (resolver, NULL);
+  ensure_avahi_resolver (resolver);
 
   resolvers = g_list_prepend (resolvers, resolver);
 
@@ -1078,7 +1066,7 @@ service_resolver_changed (GVfsDnsSdResolver *resolver,
   if (resolver->is_resolved)
     {
       g_simple_async_result_set_op_res_gboolean (data->simple, TRUE);
-      g_simple_async_result_complete (data->simple);
+      g_simple_async_result_complete_in_idle (data->simple);
       resolve_data_free (data);
     }
   else
@@ -1101,7 +1089,7 @@ service_resolver_changed (GVfsDnsSdResolver *resolver,
                                            data->resolver->service_type,
                                            data->resolver->service_name,
                                            data->resolver->domain);
-          g_simple_async_result_complete (data->simple);
+          g_simple_async_result_complete_in_idle (data->simple);
           resolve_data_free (data);
         }
     }
@@ -1146,7 +1134,7 @@ service_resolver_timed_out (ResolveData *data)
                                        data->resolver->domain);
     }
 
-  g_simple_async_result_complete (data->simple);
+  g_simple_async_result_complete_in_idle (data->simple);
   data->timeout_id = 0;
   resolve_data_free (data);
   return FALSE;
@@ -1175,7 +1163,6 @@ g_vfs_dns_sd_resolver_resolve (GVfsDnsSdResolver  *resolver,
 {
   ResolveData *data;
   GSimpleAsyncResult *simple;
-  GError *error;
 
   g_return_if_fail (G_VFS_IS_DNS_SD_RESOLVER (resolver));
 
@@ -1188,20 +1175,12 @@ g_vfs_dns_sd_resolver_resolve (GVfsDnsSdResolver  *resolver,
   if (resolver->is_resolved)
     {
       g_simple_async_result_set_op_res_gboolean (simple, TRUE);
-      g_simple_async_result_complete (simple);
+      g_simple_async_result_complete_in_idle (simple);
       g_object_unref (simple);
       goto out;
     }
 
-  error = NULL;
-  if (!ensure_avahi_resolver (resolver, &error))
-    {
-      g_simple_async_result_set_from_error (simple, error);
-      g_simple_async_result_complete (simple);
-      g_object_unref (simple);
-      g_error_free (error);
-      goto out;
-    }
+  ensure_avahi_resolver (resolver);
 
   data = g_new0 (ResolveData, 1);
   data->resolver = resolver;
@@ -1222,7 +1201,9 @@ g_vfs_dns_sd_resolver_resolve (GVfsDnsSdResolver  *resolver,
 
 typedef struct
 {
-  GMainLoop *loop;
+  GMutex mutex;
+  GCond cond;
+  gboolean done;
   GError *error;
   gboolean ret;
 } ResolveDataSync;
@@ -1235,10 +1216,13 @@ resolve_sync_cb (GVfsDnsSdResolver *resolver,
   data->ret = g_vfs_dns_sd_resolver_resolve_finish (resolver,
                                                     res,
                                                     &(data->error));
-  g_main_loop_quit (data->loop);
+  g_mutex_lock (&data->mutex);
+  data->done = TRUE;
+  g_cond_signal (&data->cond);
+  g_mutex_unlock (&data->mutex);
 }
 
-
+/* Do not call from the global main loop thread. */
 gboolean
 g_vfs_dns_sd_resolver_resolve_sync (GVfsDnsSdResolver  *resolver,
                                     GCancellable       *cancellable,
@@ -1249,29 +1233,26 @@ g_vfs_dns_sd_resolver_resolve_sync (GVfsDnsSdResolver  *resolver,
 
   g_return_val_if_fail (G_VFS_IS_DNS_SD_RESOLVER (resolver), FALSE);
 
-  /* TODO: get rid of this nested mainloop, port to avahi mainloop instead  */
-  /*       see http://bugzilla.gnome.org/show_bug.cgi?id=555436#c32 */
-
   data = g_new0 (ResolveDataSync, 1);
-  /* mark the main loop as running to have an indication
-     whether g_main_loop_quit() was called before g_main_loop_run() */
-  data->loop = g_main_loop_new (NULL, TRUE);
+  g_cond_init (&data->cond);
+  g_mutex_init (&data->mutex);
 
+  g_mutex_lock (&data->mutex);
   g_vfs_dns_sd_resolver_resolve (resolver,
                                  cancellable,
                                  (GAsyncReadyCallback) resolve_sync_cb,
                                  data);
 
-  /* start main loop only if wasn't quit before
-     (i.e. in case when pulling record from cache) */
-  if (g_main_loop_is_running (data->loop))
-    g_main_loop_run (data->loop);
+  while (!data->done)
+    g_cond_wait (&data->cond, &data->mutex);
+  g_mutex_unlock (&data->mutex);
 
   ret = data->ret;
   if (data->error != NULL)
     g_propagate_error (error, data->error);
 
-  g_main_loop_unref (data->loop);
+  g_mutex_clear (&data->mutex);
+  g_cond_clear (&data->cond);
   g_free (data);
 
   return ret;
